/*
** Module   : ControlConfig.java
** Abstract : stores common non-container widget attributes
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050714   @21692 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050728   @21909 Moving screenValue variable to ControlConfig
**                           from ControlSetConfig because not only set
**                           like widgets must use it, but text related
**                           too. For example - FillIn.
** 003 EVL 20050822   @22372 Adding the event list the server will wait
**                           for.
** 004 SIY 20050831   @22495 Refactoring (see @@22394-22409) 
** 005 SIY 20050907   @22627 Fixed use of @inheritDoc tag in javadoc
** 006 SIY 20050909   @22662 Fixed setting of the label and screen value.
** 007 NVS 20051129   @23559 Changed class access to public.
** 008 SIY 20060203   @24248 Added force-label flag.
** 009 SIY 20060209   @24477 Major refactoring.
** 010 SIY 20060331   @25291 Added setLabels() method.
** 011 SIY 20060515   @26148 Encapsulated access to label field.
** 012 SIY 20060515   @26149 Moved COLUMN-LABEL attribute from 
**                           ComponentConfig.
** 013 GES 20060803   @28400 Made externalizable for better performance.
**                           Many code standard cleanups.
** 014 NVS 20060918   @29644 Added setAttrSpace(), isAttrSpace() methods
**                           that support per field level widget ATTR-SPACE
**                           attribute.
** 015 SIY 20080513   @38272 Do not allow reducing of label length.
** 016 SIY 20080504   @38547 Updated comments to reflect current findings.
** 017 GES 20090424   @41961 Import change.
** 018 SIY 20100619          Switching to own implementation of character UI.
** 019 SIY 20100804          Changes in imports, some refactorings.
** 020 SIY 20101122          Changes related to Button hierarchy cleanup.
** 021 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 022 SIY 20111202          Updated imports.
** 023 SIY 20111221          Updated imports.
** 024 HC  20140911          GUI coordinate support - data types refactoring.
** 025 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 026 HC  20141021          Implemented window move/position support.
** 027 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 028 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 029 HC  20150214          Improved positioning of side labels.
** 030 CA  20150223          When shared frames are created, the client-side label needs to be 
**                           reset; added the resetLabel for this.
** 031 HC  20150323          Removed business logic from config classes.
** 032 IAS 20150604          Moved selection-related field from the EditorConfig
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * The class holds the non-container widget generic attributes.
 */
public class ControlConfig
extends BaseConfig
{
   /** Constant identifying the selection start or end was not set. */
   public static final int INVALID_SELECTION = Integer.MIN_VALUE;

   /** The Progress AUTO-RESIZE attribute. */
   public boolean autoResize = true;
   
   /** The NO-LABELS flag. */
   public boolean labels = true;

   /** The ATTR-SPACE flag. */
   public boolean attrSpace = false;

   /** The Progress DATA-TYPE attribute. */
   public String dataType = "character";

   /** Force setting of the label. */
   public boolean forceLabel = false;
   
   /** Help string used when widget becomes focused. */
   public String help = null;
   
   /** The LABEL attribute. */
   public String label = null;

   /** The COLUMN-LABEL attribute. */
   public String clabel = null;

   /** Value of the INDEX attribute. */
   public int index = 0;

   /** Flag indicating that this widget can accept a side-label. */
   public boolean possibleSideLabel = false;

   /** When set to a positive value, it represents the ID of the side-label widget. */
   public int sideLabelId = -1;
   
   /** Flag informing the client-side the label needs to be reset. */
   public boolean resetLabel = false;
   
   /** selection is active flag */ 
   public boolean selectionActive = false;

   /** SELECTION-END Attribute */
   public int selectionEnd = INVALID_SELECTION;
   
   /** SELECTION-START Attribute */
   public int selectionStart = INVALID_SELECTION;
   
   /**
    * Default constructor (only used in de-serialization).
    */
   public ControlConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected ControlConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected ControlConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      ControlConfig cfg = (ControlConfig) config;
      
      autoResize = cfg.autoResize;
      labels = cfg.labels;
      attrSpace = cfg.attrSpace;
      dataType = cfg.dataType;
      forceLabel = cfg.forceLabel;
      help = cfg.help;
      label = cfg.label;
      clabel = cfg.clabel;
      index = cfg.index;
      possibleSideLabel = cfg.possibleSideLabel;
      sideLabelId = cfg.sideLabelId;
      resetLabel = cfg.resetLabel;
      selectionActive = cfg.selectionActive;
      selectionEnd = cfg.selectionEnd;
      selectionStart = cfg.selectionStart;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      autoResize         = in.readBoolean();
      labels             = in.readBoolean();
      attrSpace          = in.readBoolean();
      forceLabel         = in.readBoolean();
      dataType           = (String) in.readObject();
      help               = (String) in.readObject();
      label              = (String) in.readObject();
      clabel             = (String) in.readObject();
      possibleSideLabel  = in.readBoolean();
      sideLabelId        = in.readInt();
      resetLabel         = in.readBoolean();
      selectionActive     = in.readBoolean();
      selectionEnd        = in.readInt();
      selectionStart      = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeBoolean(autoResize);
      out.writeBoolean(labels);
      out.writeBoolean(attrSpace);
      out.writeBoolean(forceLabel);
      out.writeObject(dataType);
      out.writeObject(help);
      out.writeObject(label);
      out.writeObject(clabel);
      out.writeBoolean(possibleSideLabel);
      out.writeInt(sideLabelId);
      out.writeBoolean(resetLabel);
      out.writeBoolean(selectionActive);
      out.writeInt(selectionEnd);
      out.writeInt(selectionStart);
  }
}
